package com.ringlord.antbuilder;

import java.io.File;
import java.io.Writer;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;

import java.util.Map;
import java.util.HashMap;

import static com.ringlord.antbuilder.Main.VERSION;
import static com.ringlord.antbuilder.Main.VERDATE;
import static com.ringlord.antbuilder.Main.SOURCE_GENERATION;
import static com.ringlord.antbuilder.Main.EOLN;

/**
 * Based on a ridiculously large parameter list, the buildXML method
 * creates the build.xml file for Ant.
 *
 * @author K. Udo Schuermann
 **/
class BuildXMLWriter
{
  static void buildXML( final String projectName,
                        final boolean isWAR,
                        final String projectFile,
                        final String projectMain,
                        final String projectSVer,
                        final String projectTVer,
                        final String projectLibs,
                        final boolean projectImages,
                        final boolean projectSources,
                        final boolean projectEtc,
                        final boolean projectHelp,
                        final boolean projectCheckSum,
                        final boolean projectSignExecutable,
                        final boolean projectSignLibraries,
                        final String signInfoFile,
                        final String projectLog,
                        final String projectSkeleton,
                        final boolean projectJAXB,
                        final String pathSrc,
                        final String pathLib,
                        final String pathImg,
                        final String pathHelp,
                        final String pathEtc,
                        final String pathBuild )
    throws IOException
  {
    final String execExt = (isWAR ? ".war" : ".jar");

    final Writer f = new FileWriter( "_build.xml" );

    f.write( "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"+EOLN+
             "<!--\t************************************************************"+EOLN+
             "\tAnt 1.6 (or later) Project Build File for: "+projectName+EOLN+
             "\tConstructed with AntBuilder "+VERSION+" ("+VERDATE+")"+EOLN+
             "\tSee http://ringlord.com/products/antbuilder/"+EOLN+
             "\t************************************************************ -->"+EOLN+
             "<project name=\""+projectName+"\" default=\"compile\" basedir=\".\">"+EOLN+
             "\t<property name=\"build-dir\" value=\""+pathBuild+"\"/>"+EOLN );
    f.write( "\t<property name=\"src-dir\"   value=\""+pathSrc+"\"/>"+EOLN );

    final File srcDir = new File( pathSrc );
    final File libDir = new File( pathLib );
    final File imgDir = new File( pathImg );
    final File helpDir = new File( pathHelp );
    final File etcDir = new File( pathEtc );
    final File jaxbDir = new File( "jaxb" );

    if( ! srcDir.exists() &&
        ! srcDir.mkdirs() )
      {
        System.err.println( "Failed to create source directory: "+pathSrc );
      }
    if( srcDir.exists() &&
        ! SOURCE_GENERATION[0].equals(projectSkeleton) )
      {
        final String packageName;
        final int lastDot = projectMain.lastIndexOf(".");
        if( lastDot < 0 )
          {
            packageName = "net.example.project";
          }
        else
          {
            packageName = projectMain.substring( 0, lastDot );
          }

        final File topDir;
        if( packageName.length() > 0 )
          {
            topDir = new File( srcDir, packageName.replace(".","/") );
            topDir.mkdirs();
          }
        else
          {
            topDir = srcDir;
          }

        final String mainClass = projectMain.substring( lastDot+1 );

        final Map<String,String> parameters = new HashMap<String,String>();
        parameters.put( "PROJECT_NAME", projectName );
        parameters.put( "PACKAGE_NAME", packageName );
        parameters.put( "MAIN_CLASS", mainClass );

        if( SOURCE_GENERATION[1].equals(projectSkeleton) ) // Swing
          {
            final File guiDir = new File(topDir,"gui");
            guiDir.mkdirs();

            final File mainFile        = new File(topDir,mainClass+".java");
            final File guiFile         = new File(guiDir,"GUI.java");
            final File swingWorkerFile = new File(guiDir,"JListBuilder.java");

            processTemplate( "etc/swing-main.template", mainFile, parameters );
            processTemplate( "etc/swing-gui.template", guiFile, parameters );
            processTemplate( "etc/swing-jlistbuilder.template", swingWorkerFile, parameters );
          }
        else if( SOURCE_GENERATION[2].equals(projectSkeleton) ) // JavaFX
          {
            final File guiDir = new File(topDir,"gui");
            guiDir.mkdirs();

            final File mainFile = new File(topDir,mainClass+".java");
            final File guiFile  = new File(guiDir,"GUI.java");

            processTemplate( "etc/javafx-main.template", mainFile, parameters );
            processTemplate( "etc/javafx-gui.template", guiFile, parameters );
          }
      }
    if( projectLibs.length() > 0 )
      {
        f.write( "\t<property name=\"lib-dir\"   value=\""+pathLib+"\"/>"+EOLN );
        if( ! libDir.exists() &&
            ! libDir.mkdirs() )
          {
            System.err.println( "Failed to create library directory: "+pathLib );
          }
      }
    else
      {
        if( libDir.exists() &&
            ! libDir.delete() )
          {
            System.err.println( "Could not clean up unwanted library directory "+
                                "(not empty?): "+pathLib );
          }
      }
    if( projectImages )
      {
        f.write( "\t<property name=\"img-dir\"   value=\""+pathImg+"\"/>"+EOLN );
        if( ! imgDir.exists() &&
            ! imgDir.mkdirs() )
          {
            System.err.println( "Failed to create images directory: "+pathImg );
          }
      }
    else
      {
        if( imgDir.exists() &&
            ! new File( pathImg ).delete() )
          {
            System.err.println( "Could not clean up unwanted images directory "+
                                "(not empty?): "+pathImg );
          }
      }
    if( projectEtc )
      {
        f.write( "\t<property name=\"etc-dir\"   value=\""+pathEtc+"\"/>"+EOLN );
        if( ! etcDir.exists() &&
            ! etcDir.mkdirs() )
          {
            System.err.println( "Failed to create miscellaneous directory: "+pathEtc );
          }
      }
    else if( etcDir.exists() &&
             ! etcDir.delete() )
      {
        System.err.println( "Could not clean up unwanted miscellaneous directory "+
                            "( not empty?): "+pathEtc );
      }
    if( projectHelp )
      {
        f.write( "\t<property name=\"help-dir\"  value=\""+pathHelp+"\"/>"+EOLN );
        if( ! helpDir.exists() &&
            ! helpDir.mkdirs() )
          {
            System.err.println( "Failed to create help directory: "+pathHelp );
          }
      }
    else if( helpDir.exists() &&
             ! helpDir.delete() )
      {
        System.err.println( "Could not clean up unwanted help directory "+
                            "(not empty?): "+pathHelp );
      }
    if( projectJAXB )
      {
        f.write( "\t<property name=\"jaxb-dir\"  value=\"jaxb\"/>"+EOLN );
        if( ! jaxbDir.exists() &&
            ! jaxbDir.mkdirs() )
          {
            System.err.println( "Failed to create jaxb directory: jaxb" );
          }
      }
    else
      {
        if( jaxbDir.exists() &&
            ! jaxbDir.delete() )
          {
            System.err.println( "Could not clean up unwanted jaxb directory "+
                                "(not empty?): jaxb" );
          }
      }
    f.write( "\t<property name=\"jdoc-path\" value=\"jdoc\"/>"+EOLN );
    f.write( "\t<property name=\"archive\"   value=\""+projectFile+execExt+"\"/>"+EOLN );
    f.write( "\t<property name=\"tar-file\"  value=\""+projectFile+".tar\"/>"+EOLN );
    f.write( "\t<property name=\"source\"    value=\""+projectSVer+"\"/>"+EOLN );
    f.write( EOLN );

    if( projectSignExecutable )
      {
        f.write( "\t<!-- The following file must provide information about the"+EOLN+
                 "\t     keystore. Unless you really know what you're doing,"+EOLN+
                 "\t     the keystore should be completely outside the project"+EOLN+
                 "\t     directory. You do not want to distribute it, and you"+EOLN+
                 "\t     do not want to commit it into the repository. This"+EOLN+
                 "\t     information should be supplied individually by the"+EOLN+
                 "\t     person or entity building the code. If the keystore"+EOLN+
                 "\t     fell into someone else's hands, they could sign and"+EOLN+
                 "\t     distribute any and all code as if you had done so."+EOLN+
                 EOLN+
                 "\t     The following four keys must be defined (as key=value"+EOLN+
                 "\t     pairs) in the "+signInfoFile+" file:"+EOLN+
                 "\t          keystore-file   The Java Keystore File"+EOLN+
                 "\t          keystore-pass   The password for the keystore"+EOLN+
                 "\t          key-alias       The alias for the signing key"+EOLN+
                 "\t          key-pass        The password for the signing key"+EOLN+
                 "-->"+EOLN+
                 "\t<property file=\""+signInfoFile+"\"/>"+EOLN+
                 EOLN );
      }

    if( projectJAXB )
      {
        f.write( "\t<condition property=\"generated-files-are-missing\">"+EOLN+
                 "\t\t<not>"+EOLN+
                 "\t\t\t<and>"+EOLN+
                 "\t\t\t\t<available file=\"${src-dir}/generated/*.java\"/>"+EOLN+
                 "\t\t\t</and>"+EOLN+
                 "\t\t</not>"+EOLN+
                 "\t</condition>"+EOLN+
                 EOLN );
      }
    if( projectLibs.length() > 0 )
      {
        f.write( "\t<path id=\"libraries\">"+EOLN+
                 "\t\t<fileset dir=\"${lib-dir}\">"+EOLN+
                 "\t\t\t<include name=\"**/*.jar\"/>"+EOLN+
                 "\t\t\t<include name=\"**/*.zip\"/>"+EOLN+
                 "\t\t</fileset>"+EOLN+
                 "\t</path>"+EOLN+
                 EOLN );
      }

    f.write( "\t<target name=\"all\" "+
             "depends=\"clean,jar,docs\" "+
             "description=\"Cleans project and then builds the "+projectFile+execExt+
             " and the JavaDocs\"/>"+EOLN+
             EOLN );

    f.write( "\t<target name=\"clean\" "+
             "description=\"Removes files that are rebuilt during compilation\">"+EOLN+
             "\t\t<delete dir=\"${build-dir}\" verbose=\"false\" quiet=\"true\"/>"+EOLN+
             "\t\t<delete file=\".manifest\" verbose=\"false\" quiet=\"true\"/>"+EOLN );
    if( projectJAXB )
      {
        f.write( "\t\t<!--"+EOLN+
                 "\t\t\twipe out the dynamically built files, too, but"+EOLN+
                 "\t\t\tensure that they are rebuilt even if we found"+EOLN+
                 "\t\t\tthem initially and set a flag that they were"+EOLN+
                 "\t\t\tnot missing: now they are!"+EOLN+
                 "\t\t-->"+EOLN+
                 "\t\t<delete dir=\"${src-dir}/generated\" verbose=\"false\" quiet=\"true\"/>"+EOLN+
                 "\t\t<property name=\"generated-files-are-missing\" value=\"\"/>"+EOLN );
      }
    f.write( "\t</target>"+EOLN+
             EOLN );

    f.write( "\t<target name=\"cleanall\" "+
             "depends=\"clean\" "+
             "description=\"Cleans GNU/Emacs backup files, too\">"+EOLN+
             "\t\t<delete dir=\"${jdoc-path}\" verbose=\"false\" quiet=\"true\"/>"+EOLN+
             "\t\t<delete><!-- cleanup gnuemacs backup files -->"+EOLN+
             "\t\t\t<fileset dir=\".\" defaultexcludes=\"false\">"+EOLN+
             "\t\t\t\t<include name=\"*~\"/>"+EOLN+
             "\t\t\t</fileset>"+EOLN+
             "\t\t\t<fileset dir=\"${src-dir}\" defaultexcludes=\"false\">"+EOLN+
             "\t\t\t\t<include name=\"*~\"/>"+EOLN+
             "\t\t\t\t<include name=\"**/*~\"/>"+EOLN+
             "\t\t\t</fileset>"+EOLN );
    if( projectEtc )
      {
        f.write( "\t\t\t<fileset dir=\"${etc-dir}\" defaultexcludes=\"false\">"+EOLN+
                 "\t\t\t\t<include name=\"**/*~\"/>"+EOLN+
                 "\t\t\t</fileset>"+EOLN );
      }
    if( projectHelp )
      {
        f.write( "\t\t\t<fileset dir=\"${help-dir}\" defaultexcludes=\"false\">"+EOLN+
                 "\t\t\t\t<include name=\"**/*~\"/>"+EOLN+
                 "\t\t\t</fileset>"+EOLN );
      }
    f.write( "\t\t</delete>"+EOLN+
             "\t\t<delete file=\"${archive}\" verbose=\"false\" quiet=\"true\"/>"+EOLN+
             "\t\t<delete file=\"${tar-file}\" verbose=\"false\" quiet=\"true\"/>"+EOLN+
             "\t\t<delete file=\"${archive}.md5\" verbose=\"false\" quiet=\"true\"/>"+EOLN+
             "\t\t<delete file=\"${archive}.sha1\" verbose=\"false\" quiet=\"true\"/>"+EOLN+
             "\t\t<delete file=\"${archive}.sha256\" verbose=\"false\" quiet=\"true\"/>"+EOLN+
             "\t</target>"+EOLN+
             EOLN );

    if( projectJAXB )
      {
        f.write( "\t<target name=\"compile\" "+
                 "depends=\"jaxb-generated\" "+
                 "description=\"Compiles the project into "+pathBuild+"/\">"+EOLN );
      }
    else
      {
        f.write( "\t<target name=\"compile\" "+
                 "description=\"Compiles the project into "+pathBuild+"/\">"+EOLN );
      }
    f.write( "\t\t<property name=\"target\" value=\""+projectTVer+"\"/>"+EOLN );
    f.write( "\t\t<echo message=\"Compiling Java ${source} sources for ${target} JVM\"/>"+EOLN );
    if( projectLibs.length() > 0 )
      {
        f.write( "\t\t<mkdir  dir=\"${lib-dir}\"/>"+EOLN );
      }
    if( projectImages )
      {
        f.write( "\t\t<mkdir  dir=\"${img-dir}\"/>"+EOLN );
      }
    f.write( "\t\t<mkdir  dir=\"${build-dir}\"/>"+EOLN+
             "\t\t<javac  srcdir=\"${src-dir}\""+EOLN+
             "\t\t\tdestdir=\"${build-dir}\""+EOLN+
             "\t\t\tsource=\"${source}\""+EOLN+
             "\t\t\ttarget=\"${target}\""+EOLN );
    if( projectLibs.length() > 0 )
      {
        f.write( "\t\t\tclasspathref=\"libraries\""+EOLN );
      }
    f.write( "\t\t\tdeprecation=\"true\""+EOLN+
             "\t\t\tincludeantruntime=\"false\""+EOLN+
             "\t\t\tdebug=\"true\">"+EOLN+
             "\t\t\t<compilerarg value=\"-Xlint:all,-serial\"/>"+EOLN+
             "\t\t</javac>"+EOLN+
             "\t</target>"+EOLN+
             EOLN );

    f.write( "\t<target name=\"jar\" "+
             "depends=\"clean,compile\" "+
             "description=\"Constructs the "+projectFile+execExt+" file\">"+EOLN+
             "\t\t<tstamp>"+EOLN+
             "\t\t\t<format property=\"TODAY\" pattern=\"yyyy-MMM-dd HH:mm:ss Z\"/>"+EOLN+
             "\t\t</tstamp>"+EOLN );

    if( "reference".equals(projectLibs) )
      {
        f.write( "\t\t<pathconvert\tdirsep=\"/\""+EOLN+
                 "\t\t\t\tpathsep=\" \""+EOLN+
                 "\t\t\t\tproperty=\"jar-libs\""+EOLN+
                 "\t\t\t\trefid=\"libraries\""+EOLN+
                 "\t\t\t\tsetonempty=\"true\">"+EOLN+
                 "\t\t\t<map from=\"${user.dir}/\" to=\"\"/>"+EOLN+
                 "\t\t</pathconvert>"+EOLN );
      }

    f.write( "\t\t<manifest file=\".manifest\" mode=\"update\">"+EOLN+
             "\t\t\t<attribute name=\"Main-Class\" value=\""+projectMain+"\"/>"+EOLN );
    if( projectImages )
      {
        f.write( "\t\t\t<attribute name=\"SplashScreen-Image\" value=\"${img-dir}/splash.png\"/>"+EOLN );
      }

    if( "reference".equals(projectLibs) )
      {
        f.write( "\t\t\t<attribute name=\"Class-Path\" value=\"${jar-libs}\"/>"+EOLN );
      }

    f.write( "\t\t\t<attribute name=\"Implementation-Title\" value=\""+projectName+"\"/>"+EOLN+
             "\t\t\t<attribute name=\"Implementation-Version\" value=\"1.0 (${TODAY})\"/>"+EOLN+
             "\t\t</manifest>"+EOLN );
    if( "internalize".equals(projectLibs) )
      {
        f.write( "\t\t<unzip dest=\"${build-dir}\">"+EOLN+
                 "\t\t\t<fileset dir=\"${lib-dir}\">"+EOLN+
                 "\t\t\t\t<include name=\"**/*.zip\"/>"+EOLN+
                 "\t\t\t\t<include name=\"**/*.jar\"/>"+EOLN+
                 "\t\t\t</fileset>"+EOLN+
                 "\t\t</unzip>"+EOLN );
      }
    if( isWAR )
      {
        f.write( "\t\t<war      destfile=\"${archive}\""+EOLN+
                 "\t\t\tmanifest=\".manifest\""+EOLN+
                 "\t\t\twebxml=\"web.xml\""+EOLN+
                 "\t\t\twhenempty=\"fail\""+EOLN+
                 "\t\t\tindex=\"false\"><!-- index='false' allows external JARs to be found -->"+EOLN+
                 "\t\t\t<classes\tdir=\"${build-dir}\"/>"+EOLN );
        if( projectLibs.length() > 0 )
          {
            f.write( "\t\t\t<lib\t\tdir=\"${lib-dir}\""+EOLN );
          }
        if( projectImages )
          {
            f.write( "\t\t\t<zipfileset\tdir=\"${img-dir}\""+EOLN+
                     "\t\t\t\t\tprefix=\"${img-dir}\"/>"+EOLN );
          }
        if( projectEtc )
          {
            f.write( "\t\t\t<zipfileset\tdir=\"${etc-dir}\""+EOLN+
                     "\t\t\t\t\tprefix=\"${etc-dir}\"/>"+EOLN );
          }
        if( projectHelp )
          {
            f.write( "\t\t\t<zipfileset\tdir=\"${help-dir}\""+EOLN+
                     "\t\t\t\t\tprefix=\"${help-dir}\"/>"+EOLN );
          }
        if( projectJAXB )
          {
            f.write( "\t\t\t<zipfileset\tdir=\"${jaxb-dir}\""+EOLN+
                     "\t\t\t\t\tprefix=\"\"/>"+EOLN );
          }
        f.write( "\t\t</war>"+EOLN+
                 EOLN );
      }
    else
      {
        if( projectSources &&
            "internalize".equals(projectLibs) )
          {
            f.write( "\t\t<move file=\"${build-dir}/src\" tofile=\"${build-dir}/lib-src\"/>"+EOLN );
          }
        f.write( "\t\t<jar    jarfile=\"${archive}\""+EOLN+
                 "\t\t\tmanifest=\".manifest\""+EOLN+
                 "\t\t\twhenempty=\"fail\""+EOLN+
                 "\t\t\tindex=\"false\"><!-- index='false' allows external JARs to be found -->"+EOLN+
                 "\t\t\t<fileset dir=\"${build-dir}\" includes=\"**/*.class\"/>"+EOLN+
                 "\t\t\t<fileset dir=\".\" includes=\"*.fxml\"/>"+EOLN+
                 "\t\t\t<fileset dir=\".\" includes=\"LICENSE.TXT\"/>"+EOLN+
                 "\t\t\t<fileset dir=\".\" includes=\"README.text\"/>"+EOLN );
        if( projectSources )
          {
            f.write( "\t\t\t<fileset dir=\".\" includes=\"${src-dir}/**/*.java\"/>"+EOLN );
            f.write( "\t\t\t<fileset dir=\"${build-dir}\" includes=\"lib-src/**/*.java\"/>"+EOLN );
            f.write( "\t\t\t<fileset dir=\".\" includes=\"build.xml\"/>"+EOLN );
          }
        if( projectEtc )
          {
            f.write( "\t\t\t<fileset dir=\".\">"+EOLN+
                     "\t\t\t\t<include name=\"${etc-dir}/*\"/>"+EOLN+
                     "\t\t\t\t<include name=\"${etc-dir}/**/*\"/>"+EOLN+
                     "\t\t\t</fileset>"+EOLN );
          }
        if( projectHelp )
          {
            f.write( "\t\t\t<fileset dir=\".\">"+EOLN+
                     "\t\t\t\t<include name=\"${help-dir}\"/>"+EOLN+
                     "\t\t\t\t<include name=\"${help-dir}/**/*\"/>"+EOLN+
                     "\t\t\t</fileset>"+EOLN );
          }
        if( projectJAXB )
          {
            f.write( "\t\t\t<fileset dir=\".\">"+EOLN+
                     "\t\t\t\t<include name=\"${jaxb-dir}/**\"/>"+EOLN+
                     "\t\t\t</fileset>"+EOLN );
          }
        if( projectImages )
          {
            f.write( "\t\t\t<fileset dir=\".\">"+EOLN+
                     "\t\t\t\t<include name=\"${img-dir}/*.png\"/>"+EOLN+
                     "\t\t\t\t<include name=\"${img-dir}/*.jpeg\"/>"+EOLN+
                     "\t\t\t\t<include name=\"${img-dir}/*.gif\"/>"+EOLN+
                     "\t\t\t</fileset>"+EOLN );
          }
        f.write( "\t\t</jar>"+EOLN );
        if( projectSignExecutable )
          {
            f.write( "\t\t<signjar jar=\"${archive}\""+EOLN+
                     "\t\t\t signedjar=\".${archive}\""+EOLN+
                     "\t\t\t keystore=\"${keystore-file}\""+EOLN+
                     "\t\t\t storepass=\"${keystore-pass}\""+EOLN+
                     "\t\t\t alias=\"${key-alias}\""+EOLN+
                     "\t\t\t keypass=\"${key-pass}\""+EOLN+
                     "\t\t\t lazy=\"true\"/>"+EOLN );
            f.write( "\t\t<move file=\".${archive}\" "+
                     "tofile=\"${archive}\"/>"+EOLN );
          }
      }
    f.write( "\t\t<chmod perm=\"+x\" type=\"file\" file=\"${archive}\"/>"+EOLN );

    if( projectCheckSum )
      {
        f.write( "\t\t<checksum"+EOLN+
                 "\t\t\tfile=\"${archive}\""+EOLN+
                 "\t\t\tpattern=\"{0} *{1}\""+EOLN+
                 "\t\t\talgorithm=\"MD5\""+EOLN+
                 "\t\t\tfileext=\".md5\""+EOLN+
                 "\t\t\tforceOverwrite=\"yes\"/>"+EOLN );

        f.write( "\t\t<checksum"+EOLN+
                 "\t\t\tfile=\"${archive}\""+EOLN+
                 "\t\t\tpattern=\"{0} *{1}\""+EOLN+
                 "\t\t\talgorithm=\"SHA1\""+EOLN+
                 "\t\t\tfileext=\".sha1\""+EOLN+
                 "\t\t\tforceOverwrite=\"yes\"/>"+EOLN );

        f.write( "\t\t<checksum"+EOLN+
                 "\t\t\tfile=\"${archive}\""+EOLN+
                 "\t\t\tpattern=\"{0} *{1}\""+EOLN+
                 "\t\t\talgorithm=\"SHA-256\""+EOLN+
                 "\t\t\tfileext=\".sha256\""+EOLN+
                 "\t\t\tforceOverwrite=\"yes\"/>"+EOLN );
      }

    f.write( "\t</target>"+EOLN+
             EOLN );

    if( projectSignExecutable )
      {
        f.write( "\t<target name=\"webstart\" "+
                 "depends=\"clean,jar\" "+
                 "description=\"Constructs the "+projectFile+execExt+" and signs it"+
                 ("reference".equals(projectLibs) && projectSignLibraries
                  ? ", and all referenced libraries, too"
                  : "")+
                 "\">"+EOLN );
        f.write( "\t\t<mkdir dir=\"jws\"/>"+EOLN );
        f.write( "\t\t<copy file=\"${archive}\" "+
                 "tofile=\"jws/${archive}\"/>"+EOLN );
        if( "reference".equals(projectLibs) &&
            projectSignLibraries )
          {
            f.write( "\t\t<signjar keystore=\"${keystore-file}\""+EOLN+
                     "\t\t\t storepass=\"${keystore-pass}\""+EOLN+
                     "\t\t\t alias=\"${key-alias}\""+EOLN+
                     "\t\t\t keypass=\"${key-pass}\""+EOLN+
                     "\t\t\t destDir=\"jws\""+EOLN+
                     "\t\t\t lazy=\"false\">"+EOLN+
                     "\t\t\t<fileset dir=\"${lib-dir}\">"+EOLN+
                     "\t\t\t\t<include name=\"**/*.jar\"/>"+EOLN+
                     "\t\t\t</fileset>"+EOLN+
                     "\t\t</signjar>"+EOLN );
          }
        f.write( "\t\t<chmod perm=\"+x\" type=\"file\" file=\"jws/${archive}\"/>"+EOLN );
        f.write( "\t</target>"+EOLN+
                 EOLN );
      }
    f.write( "\t<target name=\"dist\" "+
             "depends=\"clean,jar\" "+
             "description=\"Builds a distribution in "+projectFile+".tar.gz"+
             (projectSources
              ? " INCLUDING source code"
              : "")+
             "\">"+EOLN+
             "\t\t<tar    tarfile=\"${tar-file}.gz\""+EOLN+
             "\t\t\tlongfile=\"gnu\""+EOLN+
             "\t\t\tcompression=\"gzip\">"+EOLN+
             "\t\t\t<tarfileset\tdir=\".\""+EOLN+
             "\t\t\t\t\tprefix=\""+projectFile+"\""+EOLN+
             "\t\t\t\t\tmode=\"750\""+EOLN+
             "\t\t\t\t\tdirmode=\"750\""+EOLN+
             "\t\t\t\t\tusername=\"${user.name}\""+EOLN+
             "\t\t\t\t\tgroup=\""+projectFile+"\">"+EOLN+
             "\t\t\t\t<include name=\"${archive}\"/>"+EOLN );
    f.write( "\t\t\t\t<include name=\""+projectFile+"\"/>"+EOLN+
             "\t\t\t\t<include name=\""+projectFile+".bat\"/>"+EOLN+
             "\t\t\t</tarfileset>"+EOLN );
    if( (projectLibs.length() > 0) ||
        projectImages ||
        projectHelp )
      {
        f.write( "\t\t\t<tarfileset\tdir=\".\""+EOLN+
                 "\t\t\t\t\tprefix=\""+projectFile+"\""+EOLN+
                 "\t\t\t\t\tmode=\"640\""+EOLN+
                 "\t\t\t\t\tdirmode=\"750\""+EOLN+
                 "\t\t\t\t\tusername=\"${user.name}\""+EOLN+
                 "\t\t\t\t\tgroup=\""+projectFile+"\">"+EOLN );
        if( projectLibs.length() > 0 )
          {
            f.write( "\t\t\t\t<include name=\"${lib-dir}/**\"/>"+EOLN );
          }
        if( projectImages )
          {
            f.write( "\t\t\t\t<include name=\"${img-dir}/**\"/>"+EOLN );
          }
        if( projectHelp )
          {
            f.write( "\t\t\t\t<include name=\"${help-dir}/*\"/>"+EOLN+
                     "\t\t\t\t<include name=\"${help-dir}/**/*\"/>"+EOLN );
          }
        if( projectJAXB )
          {
            f.write( "\t\t\t\t<include name=\"${jaxb-dir}/**\"/>"+EOLN );
          }
        f.write( "\t\t\t</tarfileset>"+EOLN );
      }
    f.write( "\t\t\t<tarfileset\tdir=\".\""+EOLN+
             "\t\t\t\t\tprefix=\""+projectFile+"\""+EOLN+
             "\t\t\t\t\tmode=\"640\""+EOLN+
             "\t\t\t\t\tdirmode=\"750\""+EOLN+
             "\t\t\t\t\tusername=\"${user.name}\""+EOLN+
             "\t\t\t\t\tgroup=\""+projectFile+"\">"+EOLN+
             "\t\t\t\t<include name=\"README.text\"/>"+EOLN+
             "\t\t\t\t<include name=\"*.od?\"/>"+EOLN+
             "\t\t\t\t<include name=\"*.pdf\"/>"+EOLN );
    if( projectSources )
      {
        f.write( "\t\t\t\t<include name=\"build.xml\"/>"+EOLN+
                 "\t\t\t\t<include name=\"${src-dir}/*.java\"/>"+EOLN+
                 "\t\t\t\t<include name=\"${src-dir}/**/*.java\"/>"+EOLN );
      }
    f.write( "\t\t\t</tarfileset>"+EOLN+
             "\t\t</tar>"+EOLN+
             "\t</target>"+EOLN+
             EOLN );

    f.write( "\t<target name=\"docs\" "+
             "description=\"Builds the JavaDoc in jdoc/\">"+EOLN+
             "\t\t<delete dir=\"${jdoc-path}\" verbose=\"false\" quiet=\"true\"/>"+EOLN+
             "\t\t<mkdir  dir=\"${jdoc-path}\"/>"+EOLN+
             "\t\t<javadoc        sourcepath=\"${src-dir}\""+EOLN );
    if( projectLibs.length() > 0 )
      {
        f.write( "\t\t\t\tclasspathref=\"libraries\""+EOLN );
      }
    f.write( "\t\t\t\tdestdir=\"${jdoc-path}\""+EOLN+
             "\t\t\t\tprotected=\"true\""+EOLN+
             "\t\t\t\tauthor=\"true\""+EOLN+
             "\t\t\t\tversion=\"true\""+EOLN+
             "\t\t\t\tuse=\"true\""+EOLN+
             "\t\t\t\tcharset=\"UTF-8\""+EOLN+
             "\t\t\t\tsplitindex=\"false\""+EOLN+
             "\t\t\t\tbreakiterator=\"true\""+EOLN+
             "\t\t\t\tdoctitle=\"Documentation for '"+projectName+"'\""+EOLN+
             "\t\t\t\toverview=\"package.html\""+EOLN+
             "\t\t\t\twindowtitle=\""+projectName+"\">"+EOLN+
             "\t\t\t<link href=\"http://java.sun.com/j2se/1.5.0/docs/api/\"/>"+EOLN+
             "\t\t\t<fileset dir=\"${src-dir}\">"+EOLN+
             "\t\t\t\t<include name=\"*.java\"/>"+EOLN+
             "\t\t\t\t<include name=\"**/*.java\"/>"+EOLN+
             "\t\t\t</fileset>"+EOLN+
             "\t\t</javadoc>"+EOLN+
             "\t</target>"+EOLN+
             EOLN );

    if( "git".equals(projectLog) )
      {
        f.write( "\t<target name=\"log\" "+
                 "description=\"Generate the "+projectFile+".log from git's repository\">"+EOLN+
                 "\t\t<git    command=\"log\" output=\".git-output\"/>"+EOLN+
                 "\t\t<move   file=\".git-output\" tofile=\"ChangeLog\"/>"+EOLN+
                 "\t\t<fixcrlf srcdir=\".\" includes=\"ChangeLog\" eol=\"crlf\" eof=\"remove\"/>"+EOLN+
                 "\t\t<echo   message=\"Created file 'ChangeLog'\"/>"+EOLN+
                 "\t</target>"+EOLN+
                 "\t<macrodef name=\"git\">"+EOLN+
                 "\t\t<attribute name=\"command\"/>"+EOLN+
                 "\t\t<attribute name=\"dir\" default=\"\"/>"+EOLN+
                 "\t\t<attribute name=\"output\" default=\".git-output\"/>"+EOLN+
                 "\t\t<element name=\"args\" optional=\"true\"/>"+EOLN+
                 "\t\t<sequential>"+EOLN+
                 "\t\t\t<echo message=\"git @{command}\"/>"+EOLN+
                 "\t\t\t<exec executable=\"git\" dir=\"@{dir}\" output=\"@{output}\">"+EOLN+
                 "\t\t\t\t<arg value=\"@{command}\"/>"+EOLN+
                 "\t\t\t\t<arg value=\"-M\"/>"+EOLN+
                 "\t\t\t\t<arg value=\"--name-status\"/>"+EOLN+
                 "\t\t\t\t<arg value=\"--graph\"/>"+EOLN+
                 "\t\t\t\t<arg value=\"--full-history\"/>"+EOLN+
                 "\t\t\t\t<args/>"+EOLN+
                 "\t\t\t</exec>"+EOLN+
                 "\t\t</sequential>"+EOLN+
                 "\t</macrodef>"+EOLN );
      }
    else if( "svn".equals(projectLog) )
      {
        f.write( "\t<target name=\"log\" "+
                 "description=\"Generate the "+projectFile+".log from svn's repository\">"+EOLN+
                 "\t\t<svn    command=\"log\" output=\".svn-output\"/>"+EOLN+
                 "\t\t<move   file=\".svn-output\" tofile=\"ChangeLog\"/>"+EOLN+
                 "\t\t<fixcrlf srcdir=\".\" includes=\"ChangeLog\" eol=\"crlf\" eof=\"remove\"/>"+EOLN+
                 "\t\t<echo   message=\"Created file 'ChangeLog'\"/>"+EOLN+
                 "\t</target>"+EOLN+
                 "\t<macrodef name=\"svn\">"+EOLN+
                 "\t\t<attribute name=\"command\"/>"+EOLN+
                 "\t\t<attribute name=\"dir\" default=\"\"/>"+EOLN+
                 "\t\t<attribute name=\"output\" default=\".svn-output\"/>"+EOLN+
                 "\t\t<element name=\"args\" optional=\"true\"/>"+EOLN+
                 "\t\t<sequential>"+EOLN+
                 "\t\t\t<echo message=\"svn @{command}\"/>"+EOLN+
                 "\t\t\t<exec executable=\"svn\" dir=\"@{dir}\" output=\"@{output}\">"+EOLN+
                 "\t\t\t\t<arg value=\"@{command}\"/>"+EOLN+
                 "\t\t\t\t<arg value=\"-v\"/>"+EOLN+
                 "\t\t\t\t<arg value=\"-r\"/>"+EOLN+
                 "\t\t\t\t<arg value=\"HEAD:1\"/>"+EOLN+
                 "\t\t\t\t<args/>"+EOLN+
                 "\t\t\t</exec>"+EOLN+
                 "\t\t</sequential>"+EOLN+
                 "\t</macrodef>"+EOLN );
      }

    if( projectJAXB )
      {
        f.write( "\t<target name=\"jaxb-generated\" if=\"generated-files-are-missing\">"+EOLN+
                 "\t\t<exec     executable=\"xjc\" failifexecutionfails=\"false\">"+EOLN+
                 "\t\t\t<arg value=\"${jaxb-dir}/nodeList.xsd\"/>"+EOLN+
                 "\t\t\t<arg value=\"-d\"/><arg value=\"${src-dir}\"/>"+EOLN+
                 "\t\t\t<arg value=\"-p\"/><arg value=\"generated\"/>"+EOLN+
                 "\t\t</exec>"+EOLN+
                 "\t</target>"+EOLN+
                 EOLN );
      }

    if( "1.4".compareTo(projectSVer) >= 0 )
      {
        f.write( "\t<target name=\"jar4\" "+
                 "description=\"Builds the "+projectFile+execExt+" for Java 1.4\">"+EOLN+
                 "\t\t<property name=\"target\" value=\"1.4\"/>"+EOLN );
        f.write( "\t\t<antcall target=\"jar\"/>"+EOLN+
                 "\t</target>"+EOLN+
                 EOLN );
      }

    if( "1.5".compareTo(projectSVer) >= 0 )
      {
        f.write( "\t<target name=\"jar5\" "+
                 "description=\"Builds the "+projectFile+execExt+" for Java 1.5\">"+EOLN+
                 "\t\t<property name=\"target\" value=\"1.5\"/>"+EOLN );
        f.write( "\t\t<antcall target=\"jar\"/>"+EOLN+
                 "\t</target>"+EOLN+
                 EOLN );
      }

    if( "1.6".compareTo(projectSVer) >= 0 )
      {
        f.write( "\t<target name=\"jar6\" "+
                 "description=\"Builds the "+projectFile+execExt+" for Java 1.6\">"+EOLN+
                 "\t\t<property name=\"target\" value=\"1.6\"/>"+EOLN );
        f.write( "\t\t<antcall target=\"jar\"/>"+EOLN+
                 "\t</target>"+EOLN+
                 EOLN );
      }

    if( "1.7".compareTo(projectSVer) >= 0 )
      {
        f.write( "\t<target name=\"jar7\" "+
                 "description=\"Builds the "+projectFile+execExt+" for Java 1.7\">"+EOLN+
                 "\t\t<property name=\"target\" value=\"1.7\"/>"+EOLN );
        f.write( "\t\t<antcall target=\"jar\"/>"+EOLN+
                 "\t</target>"+EOLN+
                 EOLN );
      }

    if( "1.8".compareTo(projectSVer) >= 0 )
      {
        f.write( "\t<target name=\"jar8\" "+
                 "description=\"Builds the "+projectFile+execExt+" for Java 1.8\">"+EOLN+
                 "\t\t<property name=\"target\" value=\"1.8\"/>"+EOLN );
        f.write( "\t\t<antcall target=\"jar\"/>"+EOLN+
                 "\t</target>"+EOLN+
                 EOLN );
      }

    f.write( "</project>"+EOLN );

    f.flush();
    f.close();
  }

  private static void processTemplate( final String templateName,
                                       final File outputFile,
                                       final Map<String,String> parameters )
  {
    if( ! outputFile.exists() )
      {
        PrintWriter out = null;
        BufferedReader in = null;
        try
          {
            final ClassLoader cl = Main.class.getClassLoader();
            in = new BufferedReader( new InputStreamReader(cl.getResourceAsStream(templateName)) );

            out = new PrintWriter( outputFile );

            String s;
            while( (s = in.readLine()) != null )
              {
                s = performReplacements( s, parameters );
                out.println( s );
              }

            out.flush();
          }
        catch( NullPointerException x )
          {
            System.err.println( "WARNING: Failed to find template: "+templateName );
          }
        catch( IOException x )
          {
            System.err.println( "WARNING: Failed to create skeleton source file: "+x );
          }
        finally
          {
            if( out != null )
              {
                out.close();
              }
            if( in != null )
              {
                try
                  {
                    in.close();
                  }
                catch( IOException x )
                  {
                    // do nothing
                  }
              }
          }
      }
  }

  private static String performReplacements( String s,
                                             final Map<String,String> parameters )
  {
    int pos;
    while( (pos = s.indexOf("${")) > -1 )
      {
        final int end = s.indexOf("}",pos+2);
        if( end < 0 )
          {
            break;
          }
        final String key = s.substring(pos+2,end);
        final String val = parameters.get( key );
        s =
          s.substring(0,pos) +
          (val == null
           ? ""
           : val) +
          s.substring(end+1);
      }
    return s;
  }
}
